//                              -*- Mode: C++ -*- 
// Definition.cc --- Definition sub-hierarchy
// Author          : Nicolas Anquetil
// Created On      : Mon Dec  2 12:17:05 1996
// 


# include "Definition.hh"


// **************************************************
//					  Definition
// **************************************************

Definition::Definition(char *s)
//--------------------------------------------------
{
  Name *n;

# ifdef DEBUG
  printf("Definition::Definition (%s)\n", s);
# endif
  n = the_DB->getName(s);
  _definitionName = n;
  n->addNamedDefinition(this);
  the_DB->incCount(classType());
}  /* Definition::Definition */


# ifdef DEBUG
void Definition::hello()
//--------------------------------------------------
{
  printf("Definition (%s)\nis a ", definitionName()->string());
  SoftwareObject::hello();
}  /* Definition::hello */
# endif


// **************************************************
//					  TypedDefinition
// **************************************************

TypedDefinition::TypedDefinition(char *name, TypeUseExistence *ofType)
//--------------------------------------------------
  : Definition(name)
{
# ifdef DEBUG
  printf("TypedDefinition::TypedDefinition (%s)\n",name);
# endif
  _type = ofType;

  if (ofType != NULL)
	ofType->addUseType( this);
  the_DB->incCount(classType());
}  /* TypedDefinition::TypedDefinition */


VoidIterator<SoftwareObject*> *TypedDefinition::refersTo(class_t typeFilter)
//--------------------------------------------------
// may refer to type.
{
  if (type() == NULL)
	return( new VoidIterator<SoftwareObject*>());
  else
	return( new TypeFilterIterator<TypeUseExistence*,SoftwareObject*>(
	          new SingletonIterator<TypeUseExistence*>( type()),
			  typeFilter) );
}  /* TypedDefinition::refersTo */


# ifdef DEBUG
void TypedDefinition::hello()
//--------------------------------------------------
{
  printf("TypedDefinition (%s)\nis a ", definitionName()->string());
  Definition::hello();
  printf("\tof type (%s)\n", (type()==NULL?"NONE":type()->referenceName()->string()));
}  /* TypedDefinition::hello */
# endif


// **************************************************
//					StandaloneDefinition
// **************************************************

StandaloneDefinition::StandaloneDefinition(char *name,
//--------------------------------------------------
    TypeUseExistence *ofType, int startPos,	int endPos,
	SourceUnit *containerToBe)
  : TypedDefinition(name, ofType)
{
# ifdef DEBUG
  printf("StandaloneDefinition::StandaloneDefinition (%s)\n", name);
# endif
  _startChar = startPos;
  _endChar   = endPos;
  _containingSource = containerToBe;

  if (containerToBe != NULL)
	containerToBe->addDefinition( this);
  // if `ofType' and `containerToBe' are both non NULL, some other
  // function has to put `containerToBe' in `foundInSource' of
  // `ofType' and `ofType' in `refersToData' of `containerToBe'
  the_DB->incCount(classType());
}  /* StandaloneDefinition::StandaloneDefinition */


VoidIterator<SoftwareObject*> *StandaloneDefinition::definedBy(class_t type)
//--------------------------------------------------
// May be defined by file or routine
{
  if (containingSource() == NULL)
	return(new VoidIterator<SoftwareObject*>());
  else
	return( new TypeFilterIterator<SourceUnit*,SoftwareObject*>(
              new SingletonIterator<SourceUnit*>( containingSource()),
			  type) );
}  /* StandaloneDefinition::definedBy */


# ifdef DEBUG
void StandaloneDefinition::hello()
//--------------------------------------------------
{
  printf("StandaloneDefinition (%s)\nis a ", definitionName()->string());
  TypedDefinition::hello();
  printf("\tbeg char:%d, end char:%d.\n", startChar(), endChar());
  printf("containing source ----------\n");
  containingSource()->hello();
  printf("---------- containing source\n");
}  /* StandaloneDefinition::hello */
# endif


// **************************************************
//					  Field
// **************************************************

Field::Field(char *name, TypeUseExistence *ofType, RecordTypeDef *record)
//--------------------------------------------------
  : TypedDefinition(name, ofType)
{
# ifdef DEBUG
  printf("Field::Field(%s)\n", name);
# endif
  _isFieldMemberOf = record;

  record->addFieldMember( this);   // record should never be NULL
  the_DB->incCount(classType());
}  /* Field::Field */


VoidIterator<SoftwareObject*> *Field::definedBy(class_t type)
//--------------------------------------------------
// May be defined by record-type
{
  if (isFieldMemberOf() == NULL)
	return( new VoidIterator<SoftwareObject*>());
  else
	return( new TypeFilterIterator<RecordTypeDef*,SoftwareObject*>(
              new SingletonIterator<RecordTypeDef*>( isFieldMemberOf()),
			  type) );
}  /* Field::definedBy */


# ifdef DEBUG
void Field::hello()
//--------------------------------------------------
{
  printf("Field (%s)\nis a ", definitionName()->string());
  TypedDefinition::hello();
  printf("record type ----------\n");
  isFieldMemberOf()->hello();
  printf("---------- record type\n");
}  /* Field::hello */
# endif


// **************************************************
//					  TypeDef
// **************************************************

# ifdef DEBUG
void TypeDef::hello()
//--------------------------------------------------
{
  printf("TypeDef (%s)\nis a ", definitionName()->string());
  StandaloneDefinition::hello();
}  /* TypeDef::hello */
# endif


// **************************************************
//					  DatumDef
// **************************************************

DatumDef::DatumDef(char *name, TypeUseExistence *ofType,
//--------------------------------------------------
				   int startPos, int endPos, SourceUnit *containerToBe,
				   bool const_p, bool formalArg_p)
  : StandaloneDefinition(name, ofType, startPos, endPos, containerToBe)
{
# ifdef DEBUG
  printf("DatumDef::DatumDef (%s)\n", name);
# endif
  _isConst = const_p;
  _isFormalArgument = formalArg_p;

  /* this is already done in SoftwareObject, but for some reason it
  /* does not work.  We should find a better solution that this */
  the_DB->incCount( classType());
};  /* DatumDef::DatumDef */


# ifdef DEBUG
void DatumDef::hello()
//--------------------------------------------------
{
  printf("DatumDef (%s)\nis a ", definitionName()->string());
  StandaloneDefinition::hello();
  printf("\tconstant:%c, formal arg:%c\n",
		 isConst()?'Y':'N', isFormalArgument()?'Y':'N');
}  /* DatumDef::hello */
# endif


// **************************************************
//					EnumerationConst
// **************************************************

EnumerationConst::EnumerationConst(char *name,
//--------------------------------------------------
	EnumeratedTypeDef *enumeratedType)
  : Definition(name)
{
# ifdef DEBUG
  printf("EnumerationConst::EnumerationConst(%s)\n",name);
# endif
  _isEnumerationMemberOf = enumeratedType;
  enumeratedType->addEnumerationMember( this);
  // enumeratedType should never be NULL
 the_DB->incCount(classType());
}  /* EnumerationConst::EnumerationConst */;


VoidIterator<SoftwareObject*> *EnumerationConst::definedBy(class_t type)
//--------------------------------------------------
// May be defined by enumeration-type
{
  if (isEnumerationMemberOf() == NULL)
	return(new VoidIterator<SoftwareObject*>());
  else
	return( new TypeFilterIterator<EnumeratedTypeDef*,SoftwareObject*>(
              new SingletonIterator<EnumeratedTypeDef*>(
                isEnumerationMemberOf()),
			  type) );
}  /* EnumerationConst::definedBy */

# ifdef DEBUG
void EnumerationConst::hello()
//--------------------------------------------------
{
  printf("EnumerationConst (%s)\nis a ", definitionName()->string());
  Definition::hello();
  printf("member of ----------\n");
  isEnumerationMemberOf()->hello();
  printf("---------- member of\n");
}  /* EnumerationConst::hello */
# endif


// **************************************************
//					EnumeratedTypeDef
// **************************************************

VoidIterator<SoftwareObject*> *EnumeratedTypeDef::defines(class_t type)
//--------------------------------------------------
// May define enumeration-constant
{
  return( new TypeFilterIterator<EnumerationConst*,SoftwareObject*>(
            new LIterator<EnumerationConst*>( enumerationMembers()),
			type) );
}  /* EnumeratedTypeDef::defines */


# ifdef DEBUG
void EnumeratedTypeDef::hello()
//--------------------------------------------------
{
  printf("EnumeratedTypeDef (%s)\nis a ", definitionName()->string());
  TypeDef::hello();
  printf("has %d members\n", enumerationMembers()->size());
}  /* EnumeratedTypeDef::hello */
# endif


// **************************************************
//					RecordTypeDef
// **************************************************

VoidIterator<SoftwareObject*> *RecordTypeDef::defines(class_t type)
//--------------------------------------------------
// May define field
{
  return( new TypeFilterIterator<Field*,SoftwareObject*>(
            new LIterator<Field*>( fieldMembers()),
			type) );
}  /* RecordTypeDef::defines */


# ifdef DEBUG
void RecordTypeDef::hello()
//--------------------------------------------------
{
  printf("RecordTypeDef (%s)\nis a ", definitionName()->string());
  TypeDef::hello();
  printf("has %d members\n", fieldMembers()->size());
}  /* RecordTypeDef::hello */
# endif
