//                               -*- Mode: c++ -*- 
// Definition.h --- Definition sub-hierarchy
// Author          : Nicolas Anquetil
// Created On      : Mon Nov 25 12:15:39 1996
//


# ifndef DEFINITION_H
# define DEFINITION_H

# include "newDB.hh"
# include "SourceUnit.hh"
# include "ReferenceExistence.hh"


class Definition: public SoftwareObject {
// --------------------------------------------------
protected:
  Name *_definitionName;
  // converse: Name -> (List) namedDefinitions

public:
  Definition(char *s);

  Name *definitionName() const { return(_definitionName); };
  char *objectName()     const { return( definitionName()->string()); };

  class_t classType() { return(_definition); };
  bool is_a(class_t aType) {
	return( (aType ==_definition ) || SoftwareObject::is_a(aType)); };

# ifdef DEBUG
  void hello();
# endif
};


class TypedDefinition: public Definition {
// --------------------------------------------------
protected:
  TypeUseExistence *_type;
  // converse: TypeUseExistence -> (List) usedAsTypeOf

public:
  TypedDefinition(char *name, TypeUseExistence *ofType);

  TypeUseExistence *type() { return(_type); };

  virtual VoidIterator<SoftwareObject*> *refersTo(class_t type);
  class_t classType() { return(_typeddefinition); };
  bool is_a(class_t aType) {
	return( (aType ==_typeddefinition ) || Definition::is_a(aType)); };
# ifdef DEBUG
  void hello();
# endif
};


class StandaloneDefinition: public TypedDefinition {
// --------------------------------------------------
protected:
  int         _startChar, _endChar;
  SourceUnit *_containingSource;
  // converse: SourceUnit -> (List) contains

public:
  StandaloneDefinition(char *name, TypeUseExistence *ofType,
					   int startPos, int endPos, SourceUnit *containerToBe);

  int startChar() { return( _startChar); };
  int endChar() { return( _endChar); };
  SourceUnit *containingSource() { return(_containingSource); };

  VoidIterator<SoftwareObject*> *definedBy(class_t type);
  class_t classType() { return(_standalonedefinition); };
  bool is_a(class_t aType) {
	return( (aType ==_standalonedefinition ) || TypedDefinition::is_a(aType)); };
# ifdef DEBUG
  void hello();
# endif
};


class Field: public TypedDefinition {
// --------------------------------------------------
protected:
  RecordTypeDef *_isFieldMemberOf;
  // converse: RecordTypeDef -> (List) fieldMembers

public:
  Field(char *name, TypeUseExistence *ofType, RecordTypeDef *record);

  RecordTypeDef *isFieldMemberOf() { return(_isFieldMemberOf); };

  VoidIterator<SoftwareObject*> *definedBy(class_t type);
  class_t classType() { return(_field); };
  bool is_a(class_t aType) {
	return( (aType == _field) || TypedDefinition::is_a(aType)); };
# ifdef DEBUG
  void hello();
# endif
};


class TypeDef: public StandaloneDefinition {
// --------------------------------------------------
public:
  TypeDef(char *name, TypeUseExistence *ofType, int startPos, int endPos,
		  SourceUnit *containerToBe)
	: StandaloneDefinition(name, ofType, startPos, endPos, containerToBe) {
	  /* this is already done in SoftwareObject, but for some reason it
	  /* does not work.  We should find a better solution that this */
	  the_DB->incCount( classType());
  };

  class_t classType() { return(_typedef); };
  bool is_a(class_t aType) {
	return( (aType == _typedef) || StandaloneDefinition::is_a(aType)); };
# ifdef DEBUG
  void hello();
# endif
};


class DatumDef: public StandaloneDefinition {
// --------------------------------------------------
protected:
  bool _isConst;
  bool _isFormalArgument;

public:
  DatumDef(char *name, TypeUseExistence *ofType, int startPos, int endPos,
		   SourceUnit *containerToBe, bool const_p, bool formalArg_p);

  bool isConst() { return(_isConst); };
  bool isFormalArgument() { return(_isFormalArgument); };

  RoutineSource *declaredAsFormalArgsIn() {
	return(isFormalArgument() ? (RoutineSource *)containingSource()
		                      : (RoutineSource *)NULL); };
  class_t classType() { return(_datumdef); };
  bool is_a(class_t aType) {
	return( (aType == _datumdef) || StandaloneDefinition::is_a(aType)); };
# ifdef DEBUG
  void hello();
# endif
};


class EnumerationConst: public Definition {
// --------------------------------------------------
protected:
  EnumeratedTypeDef *_isEnumerationMemberOf;
  // converse: EnumeratedTypeDef -> (List) enumerationMembers

public:
  EnumerationConst(char *name, EnumeratedTypeDef *enumeratedType);

  EnumeratedTypeDef *isEnumerationMemberOf() {return(_isEnumerationMemberOf);};

  VoidIterator<SoftwareObject*> *definedBy(class_t type);
  class_t classType() { return(_enumerationconst); };
  bool is_a(class_t aType) {
	return( (aType == _enumerationconst) || Definition::is_a(aType)); };
# ifdef DEBUG
  void hello();
# endif
};


class EnumeratedTypeDef: public TypeDef {
// --------------------------------------------------
protected:
  List<EnumerationConst *> _enumerationMembers;
  // converse: EnumerationConst -> isEnumerationMemberOf

public:
  EnumeratedTypeDef(char *name,
					TypeUseExistence *ofType,
					int startPos,
					int endPos,
					SourceUnit *containerToBe)
	: TypeDef(name, ofType, startPos, endPos, containerToBe)
 {the_DB->incCount(classType());};

  List<EnumerationConst *> *enumerationMembers() {
	return(&_enumerationMembers); };

  void addEnumerationMember(EnumerationConst *member) {
	addInSet( enumerationMembers(), member); };
  VoidIterator<SoftwareObject*> *defines(class_t type);
  class_t classType() { return(_enumeratedtypedef); };
  bool is_a(class_t aType) {
	return( (aType ==_enumeratedtypedef ) || TypeDef::is_a(aType)); };

  virtual void sortLists() {
	sortTheList((List<SoftwareObject*>*)&_enumerationMembers);
	TypeDef::sortLists(); };

# ifdef DEBUG
  void hello();
# endif
};


class RecordTypeDef: public TypeDef {
// --------------------------------------------------
protected:
  List<Field *> _fieldMembers;
  // converse: Field -> isFieldMemberOf

public:
  RecordTypeDef(char *name,
				TypeUseExistence *ofType,
				int startPos,
				int endPos,
				SourceUnit *containerToBe)
	: TypeDef(name, ofType, startPos, endPos, containerToBe)
 {the_DB->incCount(classType()); };

  List<Field *> *fieldMembers() { return(&_fieldMembers); };

  void addFieldMember(Field *member) {
	addInSet( fieldMembers(), member); };
  VoidIterator<SoftwareObject*> *defines(class_t type);
  class_t classType() { return(_recordtypedef); };
  bool is_a(class_t aType) {
	return( (aType ==_recordtypedef ) || TypeDef::is_a(aType)); };

  virtual void sortLists() {
	sortTheList((List<SoftwareObject*>*)&_fieldMembers);
	TypeDef::sortLists(); };

# ifdef DEBUG
  void hello();
# endif
};

# endif
