//                              -*- Mode: C++ -*- 
// iterators.hh --- Definition of iterators
// Author          : Nicolas Anquetil
// Created On      : Fri Jan 29 19:57:59 1997
// 


# ifndef ITERATORS_H
# define ITERATORS_H

# ifdef REGEXP
extern "C" {
# include <regex.h>
}
# endif

# include "container.hh"

template <class T> class LIterator;


// **************************************************
template <class TOut>
class VoidIterator {
// Iterator empty, base class
// **************************************************
protected:
  bool _eoi;             // End Of Iteration
  VoidIterator<TOut> *_lengthIterator;
  // Keeps the future elements when computing length()
  // Actually a LIterator, but as it is not defined yet,
  // the progam would not compile

  bool listBuilt() { return(_lengthIterator != NULL); };
  inline virtual TOut computeNext() {};

public:
  VoidIterator() { _eoi = true;		   // Always empty
                   _lengthIterator = NULL; };
  virtual ~VoidIterator() {	if (listBuilt()) delete( _lengthIterator); };

  inline bool eoi() { return(_eoi); };

  inline int howManyLeft()
  // --------------------------------------------------
  {
	int count;
	List<TOut> *tmpList;

	if (listBuilt())
	  count = _lengthIterator->howManyLeft();
	else {
	  tmpList = new List<TOut>();
	  // while `_lengthIterator' is still NULL, next() works "normally"
	  for (count=0; ! eoi(); count++)
		tmpList->add( next());

	  // Now we change `_lengthIterator' and therefore next() at the same time
	  _lengthIterator = new LIterator<TOut>(tmpList);
	  ((LIterator<TOut>*)_lengthIterator)->recursiveDelete( true);
	  _eoi = _lengthIterator->eoi();
	}

	return(count);
  };  /* VoidIterator::howManyLeft() */

  inline TOut next()
  // --------------------------------------------------
  {
	TOut tmp;

	if (listBuilt()) {
	  tmp  = _lengthIterator->next();
	  _eoi = _lengthIterator->eoi();
	}
	else
	  tmp = computeNext();     // sets _eoi

	return( tmp);
  };  /* VoidIterator::next() */
};   /* class VoidIterator */


// **************************************************
template <class TIn, class TOut>
class BaseIterator: public VoidIterator<TOut> {
// Iterator on "something"
// **************************************************
protected:
  TIn _iterationBase;
  // upon termination should we also delete the _iterationBase ?
  bool _recursiveDelete;

public:
  BaseIterator(TIn base): VoidIterator<TOut>() {
  //--------------------------------------------------
	_iterationBase = base;
	_recursiveDelete = true; // by default, delete _iterationBase
  };
  ~BaseIterator() { if (recursiveDelete()) delete(_iterationBase); };

  bool recursiveDelete()       { return( _recursiveDelete); };
  void recursiveDelete(bool b) { _recursiveDelete = b; };
  void recursiveDeleteToggle() { _recursiveDelete = ! _recursiveDelete; };
};   /* class BaseIterator */


// **************************************************
template <class TOut>
class SingletonIterator: public BaseIterator<TOut,TOut> {
// Iterator on a single element, makes an iterator from any item
// **************************************************
protected:
  inline TOut computeNext() { _eoi= true;
                              return(_iterationBase); };

public:
  SingletonIterator(TOut base): BaseIterator<TOut,TOut>(base) {
  //--------------------------------------------------
	recursiveDelete( false);    // change BaseIterator's default
	_eoi = false;
  };

  inline int howManyLeft() { if (eoi())  return(0);
                             else        return(1); };
};   /* class SingletonIterator */


// **************************************************
template <class TIn, class TOut>
class CastIterator: public BaseIterator<VoidIterator<TIn>*,TOut> {
// Iterator on an iterator, cast type TIn into type TOut
// **************************************************
protected:
  inline TOut computeNext() {
  //--------------------------------------------------
	TIn tmp;
	tmp  = _iterationBase->next();
	_eoi = _iterationBase->eoi();
	return( (TOut)tmp);
  };

public:
  CastIterator(VoidIterator<TIn> *base):
  //--------------------------------------------------
	BaseIterator<VoidIterator<TIn>*,TOut>(base) {
	  _eoi = _iterationBase->eoi(); };
};   /* class CastIterator */


// **************************************************
template <class TOut>
class RewindIterator: public BaseIterator<VoidIterator<TOut>*,TOut> {
// Iterator on an iterator, remembers its elements to allow rewinding
// **************************************************
protected:
  List<TOut>      _rewindList;
  LIterator<TOut> *_rewindIterator;

  inline TOut computeNext()
  //--------------------------------------------------
  {
	TOut rtn;

	if (_rewindIterator != NULL) {
	  rtn = _rewindIterator->next();

	  if (_rewindIterator->eoi()) {
		_rewindIterator->recursiveDelete(false);
		delete(_rewindIterator);
		_rewindIterator = NULL;

		_eoi = _iterationBase->eoi();
	  }
	}
	else {
	  rtn = _iterationBase->next();
	  _rewindList.add(rtn);

	  _eoi = _iterationBase->eoi();
	}

	return(rtn);
  };    /* RewindIterator::computeNext() */

public:
  RewindIterator(VoidIterator<TOut> *base):
  //--------------------------------------------------
	BaseIterator<VoidIterator<TOut>*,TOut>(base) {
	  _eoi = _iterationBase->eoi();
	  _rewindIterator = NULL; };

  ~RewindIterator() { if ((_rewindIterator != NULL) && (recursiveDelete()))
	                    delete(_rewindIterator); };

  inline void rewind()
  //--------------------------------------------------
  {
	if (_rewindIterator != NULL) {
	  // note: does not delete the list
	  delete( _rewindIterator);
	}
	// now recreate a new iterator on _rewindList
	_rewindIterator = new LIterator<TOut>(&_rewindList);
	_rewindIterator->recursiveDelete(false);  //

	if (_rewindIterator->eoi()) {
	  delete(_rewindIterator);
	  _rewindIterator = NULL;

	  _eoi = _iterationBase->eoi();
	}
	else
	  _eoi = false;
  };    /* RewindIterator::rewind() */


  inline int howManyLeft()
  //--------------------------------------------------
  {
	int  rtn;

	if (_rewindIterator != NULL)
	  rtn = _rewindIterator->howManyLeft();
	else {
	  rtn = 0;

	  // What follows is ugly, trying to avoid it
	  if (! _eoi) {
		// The ugly thing, I see no other way to put _rewindIterator at
		// the end of the current _rewindList, i.e. in the current "state"
		for ( rewind(); ! _rewindIterator->eoi(); _rewindIterator->next() )
		  ;
		// More and more ugly, now _rewindIterator->eoi() is true
		// whereas we will add elements at the end of _rewindList.
		// _rewindIterator's _eoi will then be updated when we will ask
		// it a new element (in computeNext()).
	  }   /* if (! _eoi) */
	}   /* if (_rewindIterator != NULL) ... else */

	// counting the elements and putting them at the end of _rewindList.
	// They will be available for _rewindIterator to extract them
	for ( ; ! _iterationBase->eoi(); rtn++)
	  _rewindList.add(_iterationBase->next());

	// we did not change _eoi, whatever value it had.
  
	return(rtn);
  };  /* RewindIterator::howManyLeft() */
};  /* class RewindIterator */



// **************************************************
template <class TOut>
class AppendIterator: public BaseIterator<VoidIterator<TOut>*,TOut> {
// Iterator on two iterators, append the elements of the second at the
// end of the first one
// **************************************************
protected:
  VoidIterator<TOut> *_secondBase;

  inline TOut computeNext()
  //--------------------------------------------------
  {
	TOut tmp;

	if (! _iterationBase->eoi())
	  tmp = _iterationBase->next();
	else
	  tmp = _secondBase->next();

	_eoi = (_iterationBase->eoi() && _secondBase->eoi());

	return( tmp);
  };    /* AppendIterator::computeNext() */

public:
  AppendIterator(VoidIterator<TOut> *fst,VoidIterator<TOut> *snd)
  //--------------------------------------------------
	: BaseIterator<VoidIterator<TOut>*,TOut>(fst) {
	  _secondBase = snd;
	  _eoi = (_iterationBase->eoi() && _secondBase->eoi());
  };

  ~AppendIterator() { if (recursiveDelete()) delete(_secondBase); };
   // first base (i.e. _iterationBase) deletion is handled by ~BaseIterator
};   /* class AppendIterator */


// **************************************************
template <class TOut>
class MAppendIterator: public BaseIterator<VoidIterator<VoidIterator<TOut>*>*,TOut> {
// Appends a "list" of iterators
// **************************************************
protected:
  VoidIterator<TOut> *_currentIterator;
  List<VoidIterator<TOut>*> *_usedIterators;

  inline void nextIterator()
  //--------------------------------------------------
  {
	while ( (! eoi()) && _currentIterator->eoi() ) {
	  _usedIterators->add( _currentIterator);

	  if (_iterationBase->eoi())
		_eoi = true;
	  else
		_currentIterator = _iterationBase->next();
	}
  };    /* MAppendIterator::nextIterator() */

  inline TOut computeNext()
  //--------------------------------------------------
  {
	TOut tmp;
	tmp = _currentIterator->next();
	if (_currentIterator->eoi())
	  nextIterator();
	return( tmp);
  };    /* MAppendIterator::computeNext() */

public:
  MAppendIterator(VoidIterator<VoidIterator<TOut>*> *base)
  //--------------------------------------------------
	: BaseIterator<VoidIterator<VoidIterator<TOut>*>*,TOut>(base)
  {
	_usedIterators = new List<VoidIterator<TOut>*>();

	if (_iterationBase->eoi())
	  _eoi = true;
	else
	  _currentIterator = _iterationBase->next();

	nextIterator();     // in case the first one is empty
  };   /* MAppendIterator::MAppendIterator() */

  ~MAppendIterator()
  //--------------------------------------------------
  {
	if (recursiveDelete())
	  delete(_usedIterators);  /* ??? */
	else
	  while ( (_usedIterators->begin()) != (_usedIterators->end()) )
		/* ???? */ ;
  };   /* MAppendIterator::~MAppendIterator() */
};    /* class MAppendIterator */


// **************************************************
template <class TOut>
class LIterator: public BaseIterator<List<TOut>*,TOut> {
// Iterator on the List containers defined in container.hh
// **************************************************
protected:
  Cell<TOut> *_current;

  inline TOut computeNext()
  //--------------------------------------------------
  {
	if (_current == NULL)
	  _current = _iterationBase->begin();
	else
	  _current = _current->next();
	_eoi = (_current == _iterationBase->end());
	return( _current->val());
  };

public:
  LIterator(List<TOut> *base):
  //--------------------------------------------------
	BaseIterator<List<TOut>*,TOut>(base) {
	_current = NULL;
	recursiveDelete( false);        // change BaseIterator's default
	_eoi = (_iterationBase->begin() == NULL);
  };

  inline int LIterator<TOut>::howManyLeft()
  //--------------------------------------------------
  {
	Cell<TOut> *tmp;
	int   i;

	if (_current == NULL) {
	  tmp = _iterationBase->begin();
	  i = (tmp == NULL ? 0 : 1);
	}
	else {
	  tmp = _current;
	  i = 0;
	}
	while (tmp != _iterationBase->end()) {
	  i++;
	  tmp = tmp->next();
	}

	return(i);
  };
};   /* class LIterator */


// **************************************************
template <class TOut>
class VIterator: public BaseIterator<Vector<TOut>*,TOut> {
// Iterator on the Vector containers defined in container.hh
// **************************************************
protected:
  int _current;

  inline TOut computeNext()
  //--------------------------------------------------
  {
	_current++;  // first time _current == -1
	_eoi = (_current >= _iterationBase->size()-1);
	return( (*_iterationBase)[_current]);
  };

public:
  VIterator(Vector<TOut> *base):
  //--------------------------------------------------
	BaseIterator<Vector<TOut>*,TOut>(base) {
	_current = -1;
	recursiveDelete( false);        // change BaseIterator's default
	_eoi = (_iterationBase->size() == 0);
  };

  inline int VIterator<TOut>::howManyLeft()
  //--------------------------------------------------
  {
	return(_iterationBase->size()-_current-1);
  };
};   /* class VIterator */


// **************************************************
template <class TIn,class TOut>
class Apply0Iterator: public BaseIterator<VoidIterator<TIn>*,TOut> {
// Iterator on an iterator, modifies the elements by calling a function
// **************************************************
protected:
  // a function which takes a TIn and returns a TOut
  TOut (*_fct)(TIn);

  inline TOut computeNext()
  //--------------------------------------------------
  {
	TOut tmp;
	printf("Trace1:\n");

	tmp = (_fct)(_iterationBase->next());
	printf("Trace2:\n");

	_eoi = _iterationBase->eoi();

	return(tmp);
  };   /* Apply0Iterator::computeNext() */

public:
  Apply0Iterator(VoidIterator<TIn> *base,
				 TOut (*fct)(TIn))
  //--------------------------------------------------
	: BaseIterator<VoidIterator<TIn>*,TOut>(base) {
	  _fct = fct;
	  _eoi = _iterationBase->eoi();
  };
};   /* class Apply0Iterator */


// **************************************************
template <class TIn,class TParam,class TOut>
class Apply1Iterator: public BaseIterator<VoidIterator<TIn>*,TOut> {
// Iterator on an iterator, modifies the elements by calling a function
// **************************************************
protected:
  // a function that takes a TIn and a TParam returns a TOut
  TOut (*_fct)(TIn,TParam);
  TParam _parameter;

  inline TOut Apply1Iterator<TIn,TParam,TOut>::computeNext()
  //--------------------------------------------------
  {
	TOut tmp;
	tmp  = (_fct)(_iterationBase->next(), _parameter);
	_eoi = _iterationBase->eoi();
	return(tmp);
  };

public:
  Apply1Iterator(VoidIterator<TIn> *base,
				 TOut (*fct)(TIn,TParam),
				 TParam param)
  //--------------------------------------------------
	: BaseIterator<VoidIterator<TIn>*,TOut>(base) {
	  _fct = fct;
	  _parameter = param;
	  _eoi = _iterationBase->eoi();
  };
};    /* class Apply1Iterator */


// **************************************************
template <class TIn>
class Apply0FilterIterator: public BaseIterator<VoidIterator<TIn>*,TIn> {
// Iterator on an iterator, filters the elements  according to a
// boolean function
// **************************************************
protected:
  // the filtering function
  bool (*_fct)(TIn);
  TIn _nextComputed;

  inline void computeOneAhead()
  //--------------------------------------------------
  {
	_nextComputed = _iterationBase->next();
	while ( (! _iterationBase->eoi()) &&
			(! (_fct)(_nextComputed)) )
	  _nextComputed = _iterationBase->next();
	_eoi = _iterationBase->eoi();
  };

  inline TIn Apply0FilterIterator<TIn>::computeNext()
  //--------------------------------------------------
  {
	TIn tmp;
	tmp = _nextComputed;
	computeOneAhead();
	return(tmp);
  };

public:
  Apply0FilterIterator(VoidIterator<TIn> *base,
					   bool (*fct)(TIn))
  //--------------------------------------------------
	: BaseIterator<VoidIterator<TIn>*,TIn>(base) {
	  _fct = fct;
	  if (! (_eoi=_iterationBase->eoi()) )
		computeOneAhead();
  };
};    /* class Apply0FilterIterator */


// **************************************************
template <class TIn,class TOut>
class TypeFilterIterator: public BaseIterator<VoidIterator<TIn>*,TOut> {
// Iterator on an iterator, filters the elements according to a given type
// This one should be a sub class of Apply0FilterIterator
// **************************************************
protected:
  class_t _filteringType;
  TOut _nextComputed;

  inline void computeOneAhead()
  //--------------------------------------------------
  {
	TIn tmp;
	// assumes _iterationBase->eoi() is false
	tmp = _iterationBase->next();

	while ( (! _iterationBase->eoi()) &&
			(! tmp->is_a(_filteringType)) )
	  tmp = _iterationBase->next();

	_eoi = (! tmp->is_a(_filteringType) );
	_nextComputed = (TOut)tmp;
  };   /* TypeFilterIterator::computeOneAhead() */

  inline TOut computeNext()
  //--------------------------------------------------
  {
	TOut tmp;
	tmp = _nextComputed;
	if (! _iterationBase->eoi() )
	  computeOneAhead();   // sets _eoi
	else
	  _eoi = true;
	return(tmp);
  };   /* TypeFilterIterator::computeNext() */

public:
  TypeFilterIterator(VoidIterator<TIn> *base, class_t filter)
  //--------------------------------------------------
	: BaseIterator<VoidIterator<TIn>*,TOut>(base) {
	  _filteringType = filter;
	  if (! _iterationBase->eoi() )
		computeOneAhead();
	  else
		_eoi = true;
  };
};    /* class TypeFilterIterator */

# ifdef REGEXP
// **************************************************
template <class TIn>
class RegexpIterator: public BaseIterator<VoidIterator<TIn>*,TIn> {
// Iterator on an iterator, filters the element whose "name" matches a
// given pattern.  Code strongly based on Apply0FilterIterator
// **************************************************
protected:
  regex_t *_preg;
  TIn      _nextComputed;
  char    *(*_getStr)(TIn);  // returns the "string" of an element

  inline void computeOneAhead()
  //--------------------------------------------------
  {
	bool match;

	if (_iterationBase->eoi())
	  _eoi = true;
	else {
	  _nextComputed = _iterationBase->next();
	  match = regexec(_preg,_getStr(_nextComputed),0,NULL,0);

	  while ( (! _iterationBase->eoi()) && (match != 0) ) {
		_nextComputed = _iterationBase->next();
		match = regexec(_preg,_getStr(_nextComputed),0,NULL,0);
	  }

	  _eoi = (match != 0);
	}
  };   /* RegexpIterator::computeOneAhead() */

  inline TIn computeNext()
  //--------------------------------------------------
  {
	TIn tmp;
	tmp = _nextComputed;
	computeOneAhead();
	return(tmp);
  };

public:
  RegexpIterator(VoidIterator<TIn> *base, char *regexp, char *(*getStr)(TIn))
  //--------------------------------------------------
	: BaseIterator<VoidIterator<TIn>*,TIn>(base) {
	  int retn;

	  _getStr = getStr;

	  _preg = (regex_t*)malloc(sizeof(regex_t));
	  retn = regcomp(_preg, regexp, 0);

	  if (retn != 0)
		_eoi = true;
	  else
		if (! (_eoi = _iterationBase->eoi()) )
		  computeOneAhead();
  };   /* RegexpIterator::RegexpIterator() */

  ~RegexpIterator() { regfree(_preg); free(_preg); };
};   /* class RegexpIterator */
# endif

// **************************************************
template <class TOut>
class SortedUnionIterator: public AppendIterator<TOut> {
// Union of two _sorted_ Iterators
// **************************************************
protected:
  TOut _nextComputed1, _nextComputed2;
  bool _eoi1,          _eoi2;
  // a function which compares 2 TOut elements and return -1, 0 or 1
  int (*_cmp)(TOut,TOut);
  // a function which merges 2 TOut elements
  TOut (*_merge)(TOut,TOut);

  inline TOut computeNext()
  //--------------------------------------------------
  {
	TOut retn;
	int  cmpElts;

	if (_eoi1){
	  retn  = _nextComputed2;
	  if (! _secondBase->eoi() )    _nextComputed2 = _secondBase->next();
	  else                          _eoi2 = true;
	}
	else if (_eoi2) {
	  retn  =  _nextComputed1;
	  if (! _iterationBase->eoi() ) _nextComputed1 = _iterationBase->next();
	  else                          _eoi1 = true;
	}
	else {
	  cmpElts = (*_cmp)(_nextComputed1, _nextComputed2);

	  if (cmpElts == 0) {   /* _nextComputed1 == _nextComputed2 */
		retn = (*_merge)(_nextComputed1, _nextComputed2);
		if (! _iterationBase->eoi() ) _nextComputed1 = _iterationBase->next();
		else                          _eoi1 = true;

		if (! _secondBase->eoi() )    _nextComputed2 = _secondBase->next();
		else                          _eoi2 = true;
	  }
	  else if (cmpElts < 0) { /* _nextComputed1 < _nextComputed2 */
		retn = _nextComputed1;
		if (! _iterationBase->eoi())
		  _nextComputed1 = _iterationBase->next();
		else
		  _eoi1 = true;
	  }
	  else {                  /* _nextComputed1 > _nextComputed2 */
		retn = _nextComputed2;
		if (! _secondBase->eoi())
		  _nextComputed2 = _secondBase->next();
		else
		  _eoi2 = true;
	  }
	}

	_eoi = (_eoi1 && _eoi2);

	return( retn);
  };   /* SortedUnionIterator::computeNext() */

public:
  SortedUnionIterator(VoidIterator<TOut> *fst,VoidIterator<TOut> *snd,
					  int (*cmp)(TOut,TOut), TOut (*merge)(TOut,TOut))
  //--------------------------------------------------
	: AppendIterator<TOut>(fst,snd) {
	  _cmp   = cmp;
	  _merge = merge;

	  _eoi1  = _eoi2 = false;

	  if (! _iterationBase->eoi() ) _nextComputed1 = _iterationBase->next();
	  else                          _eoi1 = true;

	  if (! _secondBase->eoi() )    _nextComputed2 = _secondBase->next();
	  else                          _eoi2 = true;

	  _eoi = (_eoi1 && _eoi2);
  };   /* SortedUnionIterator::SortedUnionIterator() */
};    /* class SortedUnionIterator */


// **************************************************
template <class TOut>
class SortedInterIterator: public AppendIterator<TOut> {
// Intersection of two sorted Iterators.
// Code mainly based on TypeFilterIterator
// **************************************************
protected:
  TOut _nextComputed;
  // a function which compares 2 TOut elements and return -1, 0 or 1
  int (*_cmp)(TOut,TOut);
  // a function which merges 2 TOut elements
  TOut (*_merge)(TOut,TOut);

  inline void computeOneAhead()
  //--------------------------------------------------
  {
	TOut tmp1, tmp2;
	int  cmpElts;

	if (! _iterationBase->eoi()) tmp1 = _iterationBase->next();
	else                         _eoi = true;

	if (! _secondBase->eoi())    tmp2 = _secondBase->next();
	else                         _eoi = true;

	if (! _eoi) // we have two elements to compare
	  cmpElts = (*_cmp)(tmp1, tmp2);

	while ( (! _eoi) && (cmpElts != 0) ) { // tmp1 != tmp2
	  if (cmpElts < 0) {                   // tmp1 < tmp2
		if (! _iterationBase->eoi()) tmp1 = _iterationBase->next();
		else                         _eoi = true;
	  }

	  else {                               // tmp1 > tmp2
		if (! _secondBase->eoi())    tmp2 = _secondBase->next();
		else                         _eoi = true;
	  }

	  cmpElts = (*_cmp)(tmp1, tmp2);
	}

	if (! _eoi)
	  _nextComputed = (*_merge)(tmp1,tmp2);
	// else no nextComputed, it is EOI
  };   /* SortedInterIterator::computeOneAhead() */

  inline TOut computeNext()
  //--------------------------------------------------
  {
	TOut tmp;
	tmp = _nextComputed;
	computeOneAhead();   // sets _eoi
	return(tmp);
  };

public:
  SortedInterIterator(VoidIterator<TOut> *fst,VoidIterator<TOut> *snd,
					  int (*cmp)(TOut,TOut), TOut (*merge)(TOut,TOut))
  //--------------------------------------------------
	: AppendIterator<TOut>(fst,snd) {
	  _cmp   = cmp;
	  _merge = merge;

	  _eoi = false;
	  computeOneAhead();   // sets _eoi
  };
};    /* class SortedInterIterator */


// **************************************************
template <class TOut>
class SortedMinusIterator: public AppendIterator<TOut> {
// First iterator minus second one
// **************************************************
protected:
  TOut _nextComputed, _nextToSubstract;

  // a function which compares 2 TOut elements and return -1, 0 or 1
  int (*_cmp)(TOut,TOut);

  inline void computeOneAhead()
  //--------------------------------------------------
  {
	int  cmpElts;

	if (_iterationBase->eoi())
	  _eoi = true;
	else {
	  _nextComputed = _iterationBase->next();

	  if (_nextToSubstract != NULL)
		cmpElts = (*_cmp)(_nextComputed, _nextToSubstract);
	  else
		cmpElts = 1;   /* cmpElts>0 means  _nextComputed > _nextToSubstract */

	  while ( (! _iterationBase->eoi()) && (! _secondBase->eoi()) &&
			  (cmpElts >= 0) ) {
		if (cmpElts == 0)  /* _nextComputed == _nextToSubstract */
			_nextComputed = _iterationBase->next();

		_nextToSubstract = _secondBase->next();
		cmpElts = (*_cmp)(_nextComputed, _nextToSubstract);
	  }

	  if ( (cmpElts == 0) && (! _iterationBase->eoi()) )
		_nextComputed = _iterationBase->next();
	  else
		if ( (cmpElts == 0) ||
			 ( (cmpElts > 0) && (! _secondBase->eoi()) ) )
		  _eoi = true;
	}
  };   /* SortedMinusIterator::computeOneAhead() */

  inline TOut computeNext()
  //--------------------------------------------------
  {
	TOut tmp;
	tmp = _nextComputed;
	computeOneAhead();   // sets _eoi
	return(tmp);
  };

public:
  SortedMinusIterator(VoidIterator<TOut> *fst,VoidIterator<TOut> *snd,
					  int (*cmp)(TOut,TOut))
  //--------------------------------------------------
	: AppendIterator<TOut>(fst,snd) {
	  _cmp  = cmp;
	  if (! _secondBase->eoi())
		_nextToSubstract = _secondBase->next();
	  else
		_nextToSubstract = NULL;

	  computeOneAhead();
  };
};    /* class SortedMinusIterator */

# endif
