//                               -*- Mode: c++ -*- 
// newDB.hh --- Top of the new DataBase model hierarchy + several
//              global definitions (types, variables)
// Author          : Nicolas Anquetil
// Created On      : Tue Nov 26 23:06:54 1996
//


# ifndef NEWDB_H
# define NEWDB_H

# include <stdio.h>

extern "C" {
# include "cdbClasses.h"
# include "utilHash.h"
}

# include "cdbFileMem.hh"  // needed by container.hh included in iterators.hh
# include "iterators.hh"


# define DBVERSION "$Revision: 1.26 $"

// **********************************************************************
//				           A L L   C L A S S E S
// **********************************************************************

class CharSeq;
class Name;
class CharSeqFragment;

class SoftwareObject;

class SourceUnit;
class SourceFile;
class SourceWithinFile;
class RoutineSource;
class Block;

class ReferenceExistence;
class CommentTermExistence;
class SpecialCodeExistence;
class FileInclusionExistence;
class DataUseExistence;
class ManifestConstExistence;
class TypeUseExistence;
class RoutineCallExistence;

class Definition;
class TypedDefinition;
class StandaloneDefinition;
class Field;
class TypeDef;
class DatumDef;
class EnumerationConst;
class EnumeratedTypeDef;
class RecordTypeDef;

// **********************************************************************
//				           "N A M E"   C L A S S E S
// **********************************************************************

class CharSeq {
// --------------------------------------------------
protected:
  char *_str;

public:
  CharSeq(char *s = NULL) {
    _str = (char *)allocFileMem(strlen(s)+1);
    strcpy(_str, s);
  };

  inline void *operator new(size_t sz) { return(cdbAlloc(sz)); };
  inline void operator delete(void *o) { cdbFree(o); };

  char *string() const { return(_str); };
};


// two operators for that class.  Needed to use it in a map<...>.
inline bool operator <(const CharSeq& x, const CharSeq& y) {
// --------------------------------------------------
  return( strcmp( x.string(), y.string()) < 0); }

inline bool operator ==(const CharSeq& x, const CharSeq& y) {
// --------------------------------------------------
  return( strcmp( x.string(), y.string()) == 0); }


class Name: public CharSeq {
// --------------------------------------------------
protected:
  List<Definition *>         _namedDefinitions;
  // converse: Definition -> definitionName
  List<SourceFile *>         _namedFiles;
  // converse: SourceFile -> fileName
  List<RoutineSource *>      _namedRoutines;
  // converse: RoutineSource -> routineName
  List<ReferenceExistence *> _namedReferences;
  // converse: ReferenceExistence -> referenceName

public:
  Name(char *str = NULL) : CharSeq(str) { /* nothing more */ };

  List<Definition *>         *namedDefinitions(){return(&_namedDefinitions);};
  List<SourceFile *>         *namedFiles()      { return(&_namedFiles); };
  List<RoutineSource *>      *namedRoutines()   { return(&_namedRoutines); };
  List<ReferenceExistence *> *namedReferences() {return(&_namedReferences);};

  void addNamedDefinition(Definition *namedDefinition) {
	namedDefinitions()->add(namedDefinition); };
  void addNamedFile(SourceFile *namedFile) {
	namedFiles()->add(namedFile); };
  void addNamedRoutine(RoutineSource *namedRoutine) {
	namedRoutines()->add(namedRoutine); };
  void addNamedReference(ReferenceExistence *namedReference) {
	namedReferences()->add(namedReference); };
  VoidIterator<SoftwareObject*> *getObjectList(class_t classType);
  VoidIterator<SoftwareObject*> *possibleInstancesOfClass(class_t classType);
  void sortLists();
};   /* class Name */


class CharSeqFragment: public CharSeq {
// --------------------------------------------------
public:
  CharSeqFragment(char *str) : CharSeq(str) { /* nothing more */ };
};


// **********************************************************************
//				   H I E R A R C H Y   R O O T
// **********************************************************************


class SoftwareObject {
// --------------------------------------------------
protected:
  int _objectKey;

public:
  SoftwareObject();

  inline void *operator new(size_t sz) { return(cdbAlloc(sz)); };
  inline void operator delete(void *o) { cdbFree(o); };

  virtual inline char *objectName() const { return( NULL); };
          inline  int   objectKey()  const { return( _objectKey); };

  virtual  inline VoidIterator<SoftwareObject*> *defines(class_t type) {
	return( new VoidIterator<SoftwareObject*>()); };
  virtual  inline VoidIterator<SoftwareObject*> *definedBy(class_t type) {
	return( new VoidIterator<SoftwareObject*>()); };
  virtual  inline VoidIterator<SoftwareObject*> *refersTo(class_t type) {
	return( new VoidIterator<SoftwareObject*>()); };
  virtual  inline VoidIterator<SoftwareObject*> *referedBy(class_t type) {
	return( new VoidIterator<SoftwareObject*>()); };
  virtual  inline class_t classType() { return(_softwareobject); };
  // Tests if `aType' is a subclass of current class
  virtual bool is_a(class_t);

  // two methods for postProcessing
  inline  void  resetKey(int key)  { _objectKey = key; };
  virtual void sortLists()         { /* no list to sort */ };

# ifdef DEBUG
  // print the object on screen, mainly for debug purposes
  virtual void hello();
# endif
};   /* class SoftwareObject */


// **********************************************************************
//				        DATABASE
// **********************************************************************


// A single structure for the database as a whole
class Database {
// --------------------------------------------------
protected:
  char     				  *_version;
  /* We want a counter for each possible subclass of SoftwareObject.
     The array will be indexed by enumeration type class_t. */
  int                      _counts[MAX_CLASSES];
  List<RoutineSource *>   _allRoutines;
  List<SourceFile *>      _allFiles;
  Vector<SoftwareObject *> _allObjects;
  hashTable                _allNames;

public:
  Database(char *version);

  inline void *operator new(size_t sz) { return(cdbAlloc(sz)); };
  inline void operator delete(void *o) { cdbFree(o); };

  char *version();
  bool  checkVersion() { return(strcmp(_version,DBVERSION)==0); };
  int   count(class_t type) { return( _counts[type]); };
  List<RoutineSource *>   *allRoutines() { return( &_allRoutines); };
  List<SourceFile *>      *allFiles()    { return( &_allFiles); };
  Vector<SoftwareObject *> *allObjects()  { return( &_allObjects); };

  void incCount(class_t type) { _counts[type]++; };
  Name *getName(char *c);
  Name *findExistingName(char *c);
  void allNamesDo(void (func)(char *, char *), char *extraArg);

# ifdef DEBUG
  void hello();
# endif
};   /* class Database */


// get-hash-key fct used for the hashtable in the DB.
// The DB contains a pointer to this, therefore it is "sensible code"
char *dbKeyFunc(char *hashedObject);

// **********************************************************************
//				   G L O B A L   V A R I A B L E
// **********************************************************************

extern Database *the_DB;   // a single pointer to the whole DB ...

# endif
