//                              -*- Mode: C++ -*- 
// slist.hh --- single linked lists modified from STL "list.h"
// Author          : Nicolas Anquetil
// Created On      : Wed Jan  8 10:29:21 1997
// 

#ifndef LIST_H
#define LIST_H

#include <function.h>
#include <algobase.h>
#include <iterator.h>
#ifndef __GNUG__
#include <bool.h>
#endif
    
#ifndef Allocator
#define Allocator allocator
#include <defalloc.h>
#endif

#ifndef slist 
#define slist slist
#endif

template <class T>
class slist {
protected:
    typedef Allocator<void>::pointer void_pointer;
    struct list_node;
    friend list_node;
    struct list_node {
	void_pointer next;
	T data;
    };
#ifndef __GNUG__
    static Allocator<list_node> list_node_allocator;
    static Allocator<T> value_allocator;
#endif
public:      
    typedef T value_type;
    typedef Allocator<T> value_allocator_type;
    typedef Allocator<T>::pointer pointer;
    typedef Allocator<T>::reference reference;
    typedef Allocator<T>::const_reference const_reference;
    typedef Allocator<list_node> list_node_allocator_type;
    typedef Allocator<list_node>::pointer link_type;
    typedef Allocator<list_node>::size_type size_type;
    typedef Allocator<list_node>::difference_type difference_type;
protected:
#ifdef __GNUG__
    link_type get_node() { return (link_type)(::operator new(sizeof(list_node))); }
    void put_node(link_type p) { ::operator delete(p); }
#else
    size_type buffer_size() {
	return list_node_allocator.init_page_size();
    }
    struct list_node_buffer;
    friend list_node_buffer;
    struct list_node_buffer {
	void_pointer next_buffer;
	link_type buffer;
    };
public:
    typedef Allocator<list_node_buffer> buffer_allocator_type;
    typedef Allocator<list_node_buffer>::pointer buffer_pointer;     
protected:
    static Allocator<list_node_buffer> buffer_allocator;
    static buffer_pointer buffer_list;
    static link_type free_list;
    static link_type next_avail;
    static link_type last;
    void add_new_buffer() {
	buffer_pointer tmp = buffer_allocator.allocate((size_type)1);
	tmp->buffer = list_node_allocator.allocate(buffer_size());
	tmp->next_buffer = buffer_list;
	buffer_list = tmp;
	next_avail = buffer_list->buffer;
	last = next_avail + buffer_size();
    }
    static size_type number_of_lists;
    void deallocate_buffers();
    link_type get_node() {
	link_type tmp = free_list;
	return free_list ? (free_list = (link_type)(free_list->next), tmp) 
	    : (next_avail == last ? (add_new_buffer(), next_avail++) 
		: next_avail++);
	// ugly code for inlining - avoids multiple returns
    }
    void put_node(link_type p) {
	p->next = free_list;
	free_list = p;
    }
#endif

    link_type node;
    size_type length;
public:
    class iterator;
    class const_iterator;
    class iterator : public forward_iterator<T, difference_type> {
    friend class slist<T>;
    friend class const_iterator;
//  friend bool operator==(const iterator& x, const iterator& y);
    protected:
	link_type node;
	iterator(link_type x) : node(x) {}
    public:
	iterator() {}
	bool operator==(const iterator& x) const { return node == x.node; }
	reference operator*() const { return (*node).data; }
	iterator& operator++() { 
	    node = (link_type)((*node).next);
	    return *this;
	}
	iterator operator++(int) { 
	    iterator tmp = *this;
	    ++*this;
	    return tmp;
	}
    };
    class const_iterator : public forward_iterator<T, difference_type> {
    friend class slist<T>;
    protected:
	link_type node;
	const_iterator(link_type x) : node(x) {}
    public:     
	const_iterator() {}
	const_iterator(const iterator& x) : node(x.node) {}
	bool operator==(const const_iterator& x) const { return node == x.node; } 
	const_reference operator*() const { return (*node).data; }
	const_iterator& operator++() { 
	    node = (link_type)((*node).next);
	    return *this;
	}
	const_iterator operator++(int) { 
	    const_iterator tmp = *this;
	    ++*this;
	    return tmp;
	}
    };
private:
    iterator the_last_node;
public:
    slist() : length(0) {
#ifndef __GNUG__
	++number_of_lists;
#endif
	node = get_node();
	(*node).next = node;
	the_last_node = node;
    }
    iterator begin() { return node; }
    const_iterator begin() const { return node; }
    iterator end() { return(the_last_node); }
    const_iterator end() const { return(the_last_node); }
    bool empty() const { return length == 0; }
    size_type size() const { return length; }
#ifndef __GNUG__
    size_type max_size() const { return list_node_allocator.max_size(); }
#endif
    reference front() { return *begin(); }
    const_reference front() const { return *begin(); }
    void swap(slist<T>& x) {
	::swap(node, x.node);
	::swap(length, x.length);
    }
    iterator insert(iterator position, const T& x) {
	link_type tmp = get_node();
#ifdef __GNUG__
	construct(&(*tmp).data, (*position.node).data);
#else
	construct(value_allocator.address((*tmp).data), (*position.node).data);
#endif
	(*position.node).data = x;
	(*tmp).next = (*position.node).next;
 	(*position.node).next = tmp;
	if (position == the_last_node)
	  the_last_node = tmp;
	++length;
	return tmp;
    }
#ifdef __GNUG__
    void insert(iterator position, const T* first, const T* last) {
	while (first != last) insert(position, *first++);
    }
    void insert(iterator position, const_iterator first,
		const_iterator last) {
        while (first != last) insert(position, *first++);
    }
    void insert(iterator position, size_type n, const T& x) {
	while (n--) insert(position, x);
    }
#else
    void insert(iterator position, const T* first, const T* last);
    void insert(iterator position, const_iterator first,
		const_iterator last);
    void insert(iterator position, size_type n, const T& x);
#endif
    iterator insertSorted(const T& x) {
	  iterator pos = begin();
	  while ( (pos != end()) && ((*pos) < x) )
		pos++;
	  insert(pos,x);
	}
    void push_front(const T& x) { insert(begin(), x); }
    void push_back(const T& x) { insert(end(), x); }
    void erase(iterator position) {
	  iterator the_next = (link_type((*position.node).next));
 	(*position.node).next = (*the_next.node).next;
	if (the_next == the_last_node)
	  the_last_node = position;
#ifdef __GNUG__
	destroy(&(*position.node).data);
#else
	destroy(value_allocator.address((*position.node).data));
#endif
	(*position.node).data = (*the_next.node).data;
	put_node(the_next.node);
	--length;
    }
#ifdef __GNUG__
//     void erase(iterator first, iterator last) {
// 	while (first != last) erase(first++);
//     }
#else
//    void erase(iterator first, iterator last);
#endif
    void pop_front() { erase(begin()); }
    slist(size_type n, const T& value = T()) : length(0) {
#ifndef __GNUG__
	++number_of_lists;
#endif
	node = get_node();
	(*node).next = node;
	the_last_node = node;
	insert(begin(), n, value);
    }
    slist(const T* first, const T* last) : length(0) {
#ifndef __GNUG__
	++number_of_lists;
#endif
	node = get_node();
	(*node).next = node;
	the_last_node = node;
 	insert(begin(), first, last);
	}
    slist(const slist<T>& x) : length(0) {
#ifndef __GNUG__
	++number_of_lists;
#endif
	node = get_node();
	(*node).next = node;
	the_last_node = node;
	insert(begin(), x.begin(), x.end());
    }
    ~slist() {
	  while ( begin() != end() )
		erase( begin() );
	put_node(node);
#ifndef __GNUG__
	if (--number_of_lists == 0) deallocate_buffers();
#endif
    }
    slist<T>& operator=(const slist<T>& x);
#ifdef __GNUG__
    friend difference_type* distance_type(const iterator&) {
	return (difference_type*)(0);
    }
    friend T* value_type(const iterator&) {
	return (T*)(0);
    }
    friend forward_iterator_tag iterator_category(iterator) {
	return forward_iterator_tag();
    }
#endif
};

#ifndef __GNUG__
template <class T>
slist<T>::buffer_pointer slist<T>::buffer_list = 0;

template <class T>
slist<T>::link_type slist<T>::free_list = 0;

template <class T>
slist<T>::link_type slist<T>::next_avail = 0;

template <class T>
slist<T>::link_type slist<T>::last = 0;

template <class T>
slist<T>::size_type slist<T>::number_of_lists = 0;

template <class T>
slist<T>::list_node_allocator_type slist<T>::list_node_allocator;

template <class T>
slist<T>::value_allocator_type slist<T>::value_allocator;

template <class T>
slist<T>::buffer_allocator_type slist<T>::buffer_allocator;
#endif

/* 
 * currently the following does not work - made into a member function

template <class T>
inline bool operator==(const list<T>::iterator& x, const list<T>::iterator& y) { 
    return x.node == y.node; 
}
*/

template <class T>
inline bool operator==(const slist<T>& x, const slist<T>& y) {
    return x.size() == y.size() && equal(x.begin(), x.end(), y.begin());
}

template <class T>
inline bool operator<(const slist<T>& x, const slist<T>& y) {
    return lexicographical_compare(x.begin(), x.end(), y.begin(), y.end());
}

#ifndef __GNUG__
template <class T>
void slist<T>::deallocate_buffers() {
    while (buffer_list) {
	buffer_pointer tmp = buffer_list;
	buffer_list = (buffer_pointer)(buffer_list->next_buffer);
	list_node_allocator.deallocate(tmp->buffer);
	buffer_allocator.deallocate(tmp);
    }
    free_list = 0;
    next_avail = 0;
    last = 0;
}
#endif

#ifndef __GNUG__
template <class T>
void slist<T>::insert(iterator position, const T* first, const T* last) {
    while (first != last) insert(position, *first++);
}
	 
template <class T>
void slist<T>::insert(iterator position, const_iterator first,
		     const_iterator last) {
    while (first != last) insert(position, *first++);
}

template <class T>
void slist<T>::insert(iterator position, size_type n, const T& x) {
    while (n--) insert(position, x);
}

// template <class T>
// void slist<T>::erase(iterator first, iterator last) {
//     while (first != last) erase(first++);
// }
#endif

template <class T>
slist<T>& slist<T>::operator=(const slist<T>& x) {
    if (this != &x) {
	iterator first1 = begin();
	iterator last1 = end();
	const_iterator first2 = x.begin();
	const_iterator last2 = x.end();
	while (first1 != last1 && first2 != last2) *first1++ = *first2++;
	if (first2 == last2)
	    erase(first1, last1);
	else
	    insert(last1, first2, last2);
    }
    return *this;
}

#undef Allocator
#undef slist

#endif
